// SPDX-License-Identifier: LGPL-2.1-or-later
// SPDX-FileCopyrightText: 2023 Matthias Clasen
var baseURLs = [
    [ 'GLib', 'https://docs.gtk.org/glib/' ],
    [ 'GLibUnix', 'https://docs.gtk.org/glib-unix/' ],
    [ 'GLibWin32', 'https://docs.gtk.org/glib-win32/' ],
    [ 'GModule', 'https://docs.gtk.org/gmodule/' ],
    [ 'GObject', 'https://docs.gtk.org/gobject/' ],
    [ 'Gio', 'https://docs.gtk.org/gio/' ],
    [ 'GioUnix', 'https://docs.gtk.org/gio-unix/' ],
    [ 'GioWin32', 'https://docs.gtk.org/gio-win32/' ],
    [ 'Gtk', 'https://docs.gtk.org/gtk4/' ],
];
